from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Literal

from attr import Factory, define
from graphql import OperationType

if TYPE_CHECKING:
    from graphql.type import definition as gql_def
    from typing_extensions import TypeAlias

    from qtgqlcodegen.types import (
        CustomScalarDefinition,
        QtGqlEnumDefinition,
        QtGqlInputObjectTypeDefinition,
        QtGqlInterface,
        QtGqlObjectType,
        QtGqlTypeABC,
    )


@define(slots=False)
class QtGqlBaseTypedNode:
    name: str
    type: QtGqlTypeABC

    @cached_property
    def is_custom_scalar(self) -> CustomScalarDefinition | None:
        return self.type.is_custom_scalar


@define(slots=False)
class QtGqlVariableDefinition(QtGqlBaseTypedNode):
    def json_repr(self, attr_name: str | None = None) -> str:
        if not attr_name:
            attr_name = self.name
        attr_name += ".value()"  # unwrap optional
        return self.type.json_repr(attr_name)


@define(slots=False)
class BaseQtGqlFieldDefinition(QtGqlBaseTypedNode):
    description: str | None = ""


@define(slots=False)
class QtGqlArgumentDefinition(BaseQtGqlFieldDefinition, QtGqlVariableDefinition):
    ...


@define(slots=False, kw_only=True)
class QtGqlFieldDefinition(BaseQtGqlFieldDefinition):
    arguments_dict: dict[str, QtGqlArgumentDefinition] = Factory(dict)

    @cached_property
    def arguments(self) -> tuple[QtGqlArgumentDefinition, ...]:
        return tuple(self.arguments_dict.values())

    @cached_property
    def arguments_type(self) -> str:
        return "std::tuple<" + ",".join([arg.type.type_name() for arg in self.arguments]) + ">"

    def index_for_argument(self, arg: str) -> int:
        return self.arguments.index(self.arguments_dict[arg])

    @cached_property
    def getter_name(self) -> str:
        return f"get_{self.name}"

    @cached_property
    def setter_name(self) -> str:
        return f"set_{self.name}"

    @cached_property
    def signal_name(self) -> str:
        return f"{self.name}Changed"

    @cached_property
    def private_name(self) -> str:
        return f"m_{self.name}"

    @cached_property
    def implements_node(self) -> bool:
        """Helper to check whether the field type implements node."""
        object_type = self.type.is_object_type or self.type.is_interface
        if not object_type:
            raise NotImplementedError

        if object_type:
            return object_type.implements_node
        return False

    @property
    def default_value(self) -> str:
        if self.arguments:
            return "{}"
        return self.type.default_value


EnumMap: TypeAlias = "dict[str, QtGqlEnumDefinition]"
ObjectTypeMap: TypeAlias = "dict[str, QtGqlObjectType]"
InputObjectMap: TypeAlias = "dict[str, QtGqlInputObjectTypeDefinition]"
InterfacesMap: TypeAlias = "dict[str, QtGqlInterface]"
CustomScalarMap: TypeAlias = "dict[str, CustomScalarDefinition]"


@define(slots=False)
class SchemaTypeInfo:
    schema_definition: gql_def.GraphQLSchema
    custom_scalars: CustomScalarMap
    operation_types: dict[
        Literal["query", "mutation", "subscription"],
        QtGqlObjectType,
    ] = Factory(dict)
    object_types: ObjectTypeMap = Factory(dict)
    enums: EnumMap = Factory(dict)
    input_objects: InputObjectMap = Factory(dict)
    interfaces: InterfacesMap = Factory(dict)

    def get_interface(self, name: str) -> QtGqlInterface | None:
        return self.interfaces.get(name, None)

    def get_object_type(self, name: str) -> QtGqlObjectType | None:
        return self.object_types.get(name, None)

    def add_objecttype(self, objecttype: QtGqlObjectType) -> None:
        self.object_types[objecttype.name] = objecttype

    @cached_property
    def root_types(self) -> list[gql_def.GraphQLObjectType | None]:
        return [
            self.schema_definition.get_root_type(OperationType.QUERY),
            self.schema_definition.get_root_type(OperationType.MUTATION),
            self.schema_definition.get_root_type(OperationType.SUBSCRIPTION),
        ]

    @cached_property
    def root_types_names(self) -> str:
        return " ".join([tp.name for tp in self.root_types if tp])



from __future__ import annotations

from abc import ABC, abstractmethod
from functools import cached_property
from typing import TYPE_CHECKING, cast

import attrs
from attr import define

from qtgqlcodegen.core.cppref import CppAttribute, QtGqlBasesNs, QtGqlTypes

if TYPE_CHECKING:
    from qtgqlcodegen.operation.definitions import QtGqlQueriedField
    from qtgqlcodegen.schema.definitions import (
        CustomScalarMap,
        QtGqlArgumentDefinition,
        QtGqlFieldDefinition,
    )


@define
class QtGqlTypeABC(ABC):
    @property
    def is_union(self) -> QtGqlUnion | None:
        return None

    @property
    def is_optional(self) -> bool:
        return False

    @property
    def is_object_type(self) -> QtGqlObjectType | None:
        return None

    @property
    def is_interface(self) -> QtGqlInterface | None:
        return None

    @property
    def is_input_object_type(self) -> QtGqlInputObjectTypeDefinition | None:
        return None

    @property
    def is_model(self) -> QtGqlList | None:
        return None

    @property
    def is_enum(self) -> QtGqlEnumDefinition | None:
        return None

    @property
    def is_builtin_scalar(self) -> BuiltinScalar | None:
        return None

    @property
    def is_custom_scalar(self) -> CustomScalarDefinition | None:
        return None

    @property
    def is_queried_object_type(self) -> QtGqlQueriedObjectType | None:
        return None

    @property
    def is_queried_interface(self) -> QtGqlQueriedInterface | None:
        return None

    @property
    def is_queried_union(self) -> QtGqlQueriedUnion | None:
        return None

    def json_repr(self, attr_name: str | None = None) -> str:
        raise NotImplementedError(f"{self} is not supported as an input type ATM")

    @abstractmethod
    def type_name(self) -> str:
        """
        :returns: The C++ "real" type mostly this would be the member type as well.
        """
        raise NotImplementedError

    @property
    def property_type(self) -> str:
        """
        :return: The QProperty type, usually will lay in the proxy.
        """
        return f"{self.type_name()} &"  # default for scalars this should suffice.

    @property
    def default_value(self) -> str:
        """

        :return: C++ default value initializer for this type
        """
        return "{}"

    @property
    def member_type(self) -> str:
        """
        :returns: The C++ type of the concrete instance member.
        """
        # Usually `type_name` should suffice, if the type needs to return something else this is overridden.
        return self.type_name()

    @property
    def member_type_arg(self) -> str:
        """

        :return: The C++ member_type if it was passed in argument to somewhere.
        """
        return f"const {self.member_type} &"

    @property
    def fget_type(self) -> str:
        """

        :return: type for the proxy object, usually this would be the member type though for custom scalars
        there is `to_qt` that has different type.
        """
        return self.member_type

    @property
    def getter_is_constable(self) -> bool:
        return True

    def __str__(self) -> str:  # pragma: no cover
        raise RuntimeError("the template probobly tried to render this object")


@define
class QtGqlOptional(QtGqlTypeABC):
    """Represents GraphQL types that are not marked with "!"."""

    of_type: QtGqlTypeABC

    def __getattr__(self, item):
        if item == "is_optional":
            return True
        return getattr(self.of_type, item)

    def __getattribute__(self, name):
        if name not in ("of_type"):
            raise AttributeError
        else:
            return super().__getattribute__(name)

    def type_name(self) -> str:  # pragma: no cover
        raise NotImplementedError


@define
class QtGqlList(QtGqlTypeABC):
    of_type: QtGqlTypeABC

    @property
    def is_model(self) -> QtGqlList | None:
        return self

    @property
    def member_type(self) -> str:
        return f"QList<{self.of_type.member_type}>"

    def type_name(self) -> str:
        raise NotImplementedError(
            "models have no valid type for schema concretes, call member_type",
        )

    @property
    def property_type(self) -> str:
        if self.of_type.is_queried_object_type:
            return f"qtgql::bases::ListModelABC<{self.of_type.type_name()}> *"
        if self.of_type.is_queried_union:
            return f"qtgql::bases::ListModelABC<{self.of_type.is_queried_union.type_name()}> *"
        raise NotImplementedError


@define
class QtGqlUnion(QtGqlTypeABC):
    types: tuple[QtGqlObjectType | QtGqlDeferredType, ...]

    @property
    def is_union(self) -> QtGqlUnion | None:
        return self

    def type_name(self) -> str:
        return f"{QtGqlTypes.ObjectTypeABC.name}"

    @property
    def member_type(self) -> str:
        return f"std::shared_ptr<{self.type_name()}>"

    @property
    def default_value(self) -> str:
        return "{}"

    def get_by_name(self, name: str) -> QtGqlObjectType | None:
        for possible in self.types:
            if possible.name == name:
                return cast(QtGqlObjectType, possible)


@define
class BuiltinScalar(QtGqlTypeABC):
    attr: CppAttribute
    default_value_: CppAttribute
    graphql_name: str
    from_json_convertor: str

    @property
    def default_value(self) -> str:
        return self.default_value_.name

    @property
    def is_builtin_scalar(self) -> BuiltinScalar | None:
        return self

    def type_name(self) -> str:
        return self.attr.name

    @property
    def is_void(self) -> bool:
        return self is BuiltinScalars.VOID

    def json_repr(self, attr_name: str | None = None) -> str:
        return f"{attr_name}"


@define
class CustomScalarDefinition(QtGqlTypeABC):
    name: str
    graphql_name: str
    deserialized_type: str
    to_qt_type: str
    include_path: str

    @property
    def is_custom_scalar(self) -> CustomScalarDefinition | None:
        return self

    def json_repr(self, attr_name: str | None = None) -> str:
        return f"{attr_name}.serialize()"

    def type_name(self) -> str:
        return self.name

    @property
    def fget_type(self) -> str:
        return self.to_qt_type

    @property
    def getter_is_constable(self) -> bool:
        return False

    @property
    def property_type(self) -> str:
        return self.to_qt_type


@define(slots=False)
class BaseQtGqlObjectType(QtGqlTypeABC):
    name: str
    fields_dict: dict[str, QtGqlFieldDefinition]
    docstring: str | None = ""

    @cached_property
    def fields(self) -> tuple[QtGqlFieldDefinition, ...]:
        return tuple(self.fields_dict.values())


@define(slots=False)
class QtGqlObjectType(BaseQtGqlObjectType):
    interfaces_raw: tuple[QtGqlInterface, ...] = attrs.Factory(tuple)
    unique_fields: tuple[QtGqlFieldDefinition, ...] = attrs.Factory(tuple)
    is_root: bool = False

    def implements(self, interface: QtGqlInterface) -> bool:
        for m_interface in self.interfaces_raw:
            if interface is m_interface or m_interface.implements(interface):
                return True
        return False

    @cached_property
    def bases(self) -> list[QtGqlInterface]:
        """
        returns only the top level interfaces that should be inherited.
        if i.e
        ```graphql
        interface Node{
        id: ID!
        }

        interface A implements Node{
        otherField: String!
        }

        type Foo implements A{
        ...
        }
        ```
        Type `Foo` would extend only `A`

        If there are no interfaces returns only NodeInterfaceABC or ObjectTypeABC.
        """
        not_unique_interfaces: list[QtGqlInterface] = []

        if not self.interfaces_raw:
            # TODO(https://github.com/qtgql/qtgql/issues/267): these are not really interfaces though they are inherited if there are no interfaces.
            if interface := self.is_interface:
                if interface.is_node_interface:
                    return [QtGqlTypes.NodeInterfaceABC]  # type: ignore
                return [QtGqlTypes.ObjectTypeABC]  # type: ignore
            else:
                return [QtGqlTypes.ObjectTypeABC]  # type: ignore

        for interface in self.interfaces_raw:
            for other in self.interfaces_raw:
                if other is not interface:
                    if interface.implements(other):
                        not_unique_interfaces.append(other)

        return [intfs for intfs in self.interfaces_raw if intfs not in not_unique_interfaces]

    @cached_property
    def implements_node(self) -> bool:
        return any(interface.is_node_interface for interface in self.interfaces_raw)

    @property
    def is_object_type(self) -> QtGqlObjectType | None:
        return self

    def type_name(self) -> str:
        return self.name

    @property
    def member_type(self) -> str:
        if self.is_root:
            return self.type_name()  # root types are singletons
        return f"std::shared_ptr<{self.type_name()}>"

    @property
    def member_type_arg(self) -> str:
        if self.is_root:
            return f"{self.member_type} *"
        return f"const {self.member_type} &"

    def __attrs_post_init__(self):
        # inject this object type to the interface.
        # later the interface would use this list to know who he might resolve to.
        for base in self.interfaces_raw:
            if not base.implementations.get(self.name):
                base.implementations[self.name] = self


@define
class QtGqlDeferredType(QtGqlTypeABC):
    name: str
    object_map__: dict[str, QtGqlObjectType]
    cached_: QtGqlTypeABC | None = None

    def __getattr__(self, item):
        if not self.cached_:
            self.cached_ = self.object_map__[self.name]
        try:
            return getattr(self.cached_, item)
        except AttributeError:  # pragma: no cover
            raise Exception(f"{self.cached_.__class__} has no attribute named {item}")

    def __getattribute__(self, name):
        if name not in ("resolve", "object_map__", "cached_", "name"):
            raise AttributeError
        else:
            return super().__getattribute__(name)

    def type_name(self) -> str:  # we only override it since it is abstractmethod
        raise NotImplementedError("this should not be reached since we override __getattribute__")


@define(slots=False, kw_only=True)
class QtGqlInterface(QtGqlObjectType):
    implementations: dict[str, BaseQtGqlObjectType] = attrs.field(factory=dict)

    @cached_property
    def is_node_interface(self) -> bool:
        """As specified by https://graphql.org/learn/global-object-
        identification/#node-interface."""
        if self.name == "Node":
            id_field_maybe = self.fields[0]
            return id_field_maybe.name == "id" and id_field_maybe.type is BuiltinScalars.ID
        return False

    @property
    def is_object_type(self) -> QtGqlObjectType | None:
        return None  # although interface extends object this might be confusing.

    @property
    def is_interface(self) -> QtGqlInterface | None:
        return self


@define(slots=False)
class QtGqlInputObjectTypeDefinition(BaseQtGqlObjectType):
    fields_dict: dict[str, QtGqlArgumentDefinition] = attrs.field(factory=dict)  # type: ignore

    @property
    def is_input_object_type(self) -> QtGqlInputObjectTypeDefinition | None:
        return self

    def type_name(self) -> str:
        return self.name

    def json_repr(self, attr_name: str | None = None) -> str:
        return f"{attr_name}.to_json()"


@define
class EnumValue:
    """Encapsulates enumValues from introspection, maps to an Enum member."""

    name: str
    index: int
    description: str = ""


@define(slots=False)
class QtGqlEnumDefinition(QtGqlTypeABC):
    name: str
    members: list[EnumValue]

    @cached_property
    def map_name(self) -> str:
        return f"{self.name}EnumMap"

    @cached_property
    def namespaced_name(self) -> str:
        return f"Enums::{self.name}"

    @property
    def is_enum(self) -> QtGqlEnumDefinition | None:
        return self

    def type_name(self) -> str:
        return self.namespaced_name

    @property
    def default_value(self) -> str:
        return f"{self.namespaced_name}(0)"

    def json_repr(self, attr_name: str | None = None) -> str:
        return f"Enums::{self.map_name}::name_by_value({attr_name})"


@define
class QtGqlQueriedTypeABC(ABC):
    concrete: QtGqlTypeABC


@define(slots=False, repr=False)
class QtGqlQueriedObjectType(QtGqlQueriedTypeABC, QtGqlTypeABC):
    name: str
    concrete: QtGqlObjectType
    fields_dict: dict[str, QtGqlQueriedField] = attrs.Factory(dict)
    base_interface: QtGqlQueriedInterface | None = None  # I think that there could be only one

    @property
    def implements_node(self) -> bool:
        return self.concrete.implements_node

    @property
    def is_queried_object_type(self) -> QtGqlQueriedObjectType | None:
        return self

    @cached_property
    def fields(self) -> tuple[QtGqlQueriedField, ...]:
        return tuple(self.fields_dict.values())

    @property
    def deserializer_name(self) -> str:
        return f"deserializers::des_{self.name}"

    @property
    def updater_name(self) -> str:
        return f"updaters::update_{self.name}"

    def type_name(self) -> str:
        return self.name

    @property
    def property_type(self) -> str:
        return f"{self.type_name()} *"

    @cached_property
    def references(self) -> list[QtGqlQueriedField]:
        """
        :return: Fields that should be treated with special care by the operation.
        They can't just return the field of the concrete.
        """
        return [
            f
            for f in self.fields
            if (
                f.type.is_queried_object_type
                or f.type.is_queried_interface
                or f.type.is_queried_union
            )
        ]

    @cached_property
    def models(self) -> list[QtGqlQueriedField]:
        return [f for f in self.fields if f.type.is_model]

    @cached_property
    def private_name(self) -> str:
        return f"m_{self.name}"


@define(slots=False, repr=False)
class QtGqlQueriedInterface(QtGqlQueriedObjectType):
    choices: list[QtGqlQueriedObjectType] = attrs.Factory(list)

    @property
    def is_queried_object_type(self) -> QtGqlQueriedObjectType | None:
        return None

    @property
    def is_queried_interface(self) -> QtGqlQueriedInterface | None:
        return self


@define(slots=False, repr=False)
class QtGqlQueriedUnion(QtGqlQueriedTypeABC, QtGqlTypeABC):
    concrete: QtGqlUnion
    choices: tuple[QtGqlQueriedObjectType, ...]

    @property
    def is_queried_union(self) -> QtGqlQueriedUnion | None:
        return self

    def type_name(self) -> str:
        return self.concrete.type_name()

    @property
    def property_type(self) -> str:
        return f"{QtGqlTypes.ObjectTypeABC.name} *"


def ScalarsNs() -> CppAttribute:
    return QtGqlBasesNs().ns_add("scalars")


def DefaultsNs() -> CppAttribute:
    return QtGqlBasesNs().ns_add("DEFAULTS")


class _BuiltinScalars:
    INT = BuiltinScalar(
        CppAttribute("int"),
        DefaultsNs().ns_add("INT"),
        graphql_name="Int",
        from_json_convertor="toInt()",
    )
    FLOAT = BuiltinScalar(
        CppAttribute("float"),
        DefaultsNs().ns_add("FLOAT"),
        graphql_name="Float",
        from_json_convertor="toDouble()",
    )
    STRING = BuiltinScalar(
        CppAttribute("QString"),
        DefaultsNs().ns_add("STRING"),
        graphql_name="String",
        from_json_convertor="toString()",
    )
    ID = BuiltinScalar(
        ScalarsNs().ns_add("Id"),
        DefaultsNs().ns_add("ID"),
        graphql_name="ID",
        from_json_convertor="toString()",
    )
    BOOLEAN = BuiltinScalar(
        CppAttribute("bool"),
        DefaultsNs().ns_add("BOOL"),
        graphql_name="Boolean",
        from_json_convertor="toBool()",
    )
    VOID = BuiltinScalar(
        ScalarsNs().ns_add("Void"),
        DefaultsNs().ns_add("VOID"),
        graphql_name="Void",
        from_json_convertor="",
    )

    UUID = BuiltinScalar(
        CppAttribute("QUuid"),
        DefaultsNs().ns_add("UUID"),
        graphql_name="UUID",
        from_json_convertor="toVariant().toUuid()",
    )

    @classmethod
    def __iter__(cls):
        for name, member in cls.__dict__.items():
            if name.isupper():
                yield member

    def by_graphql_name(self, name: str) -> BuiltinScalar | None:
        for scalar in self:
            if scalar.graphql_name == name:
                return scalar

    @cached_property
    def keys(self) -> list[str]:
        return [scalar.graphql_name for scalar in self]


BuiltinScalars = _BuiltinScalars()
DateTimeScalarDefinition = CustomScalarDefinition(
    name="qtgql::customscalars::DateTimeScalar",
    graphql_name="DateTime",
    deserialized_type="QDateTime",
    to_qt_type="QString",
    include_path="<qtgql/customscalars/customscalars.hpp>",
)
DateScalarDefinition = CustomScalarDefinition(
    name="qtgql::customscalars::DateScalar",
    graphql_name="Date",
    deserialized_type="QDate",
    to_qt_type="QString",
    include_path="<qtgql/customscalars/customscalars.hpp>",
)
TimeScalarDefinition = CustomScalarDefinition(
    name="qtgql::customscalars::TimeScalar",
    graphql_name="Time",
    deserialized_type="QTime",
    to_qt_type="QString",
    include_path="<qtgql/customscalars/customscalars.hpp>",
)
DecimalScalarDefinition = CustomScalarDefinition(
    name="qtgql::customscalars::DecimalScalar",
    graphql_name="Decimal",
    deserialized_type="QString",
    to_qt_type="QString",
    include_path="<qtgql/customscalars/customscalars.hpp>",
)
CUSTOM_SCALARS: CustomScalarMap = {
    DateTimeScalarDefinition.graphql_name: DateTimeScalarDefinition,
    DateScalarDefinition.graphql_name: DateScalarDefinition,
    TimeScalarDefinition.graphql_name: TimeScalarDefinition,
    DecimalScalarDefinition.graphql_name: DecimalScalarDefinition,
}



from __future__ import annotations

from typing import TYPE_CHECKING

import graphql
from graphql import OperationDefinitionNode, OperationType, language as gql_lang
from graphql.language import visitor

from qtgqlcodegen.core.graphql_ref import (
    has_id_selection,
    has_typename_selection,
    inject_id_selection,
    inject_typename_selection,
    is_field_node,
    is_inline_fragment,
    is_named_type_node,
    is_nonnull_node,
    is_operation_def_node,
)
from qtgqlcodegen.operation.definitions import (
    OperationTypeInfo,
    QtGqlOperationDefinition,
    QtGqlQueriedField,
    QtGqlVariableUse,
)
from qtgqlcodegen.schema.definitions import (
    QtGqlFieldDefinition,
    QtGqlVariableDefinition,
    SchemaTypeInfo,
)
from qtgqlcodegen.schema.evaluation import evaluate_graphql_type
from qtgqlcodegen.types import (
    QtGqlInterface,
    QtGqlList,
    QtGqlOptional,
    QtGqlQueriedInterface,
    QtGqlQueriedObjectType,
    QtGqlQueriedUnion,
    QtGqlUnion,
)
from qtgqlcodegen.utils import require

if TYPE_CHECKING:
    from qtgqlcodegen.types import QtGqlObjectType, QtGqlTypeABC


def is_type_name_selection(field_node: gql_lang.FieldNode):
    # typename is not a 'real' selection and is handled with special care.
    if field_node.name.value == "__typename":
        return True
    return False


def _evaluate_variable_uses(
    type_info: OperationTypeInfo,
    field: QtGqlFieldDefinition,
    arguments: tuple[gql_lang.ArgumentNode, ...],
) -> list[QtGqlVariableUse]:
    ret: list[QtGqlVariableUse] = []
    for arg in arguments:
        index = field.index_for_argument(arg.name.value)
        var_name = arg.value.name.value  # type: ignore[attr-defined]
        for variable in type_info.variables:
            if var_name == variable.name:
                ret.append(
                    QtGqlVariableUse(argument=(index, field.arguments[index]), variable=variable),
                )
    assert len(ret) == len(arguments), "could not find all variable uses"
    return sorted(ret, key=lambda v: v.argument[0])


def _evaluate_variable_node_type(
    type_info: SchemaTypeInfo,
    node: graphql.TypeNode,
) -> QtGqlTypeABC:
    if nonnull := is_nonnull_node(node):
        return evaluate_graphql_type(
            type_info,
            graphql.type.GraphQLNonNull(
                type_info.schema_definition.get_type(nonnull.type.name.value),  # type: ignore
            ),
        )

    if named_type := is_named_type_node(node):
        gql_concrete = type_info.schema_definition.get_type(named_type.name.value)
        assert gql_concrete
        return evaluate_graphql_type(type_info, gql_concrete)
    raise NotImplementedError(node, "Type is not supported as a variable ATM")


def _evaluate_variable(
    type_info: SchemaTypeInfo,
    var: gql_lang.VariableDefinitionNode,
) -> QtGqlVariableDefinition:
    return QtGqlVariableDefinition(
        name=var.variable.name.value,
        type=_evaluate_variable_node_type(type_info, var.type),
    )


def _evaluate_selection_set_type(
    type_info: OperationTypeInfo,
    concrete_type: QtGqlTypeABC,
    selection_set_node: gql_lang.SelectionSetNode | None,
    path: str,
) -> QtGqlTypeABC:
    ret: QtGqlTypeABC | None = None
    if not selection_set_node:
        # these types have no selections
        assert (
            concrete_type.is_builtin_scalar
            or concrete_type.is_custom_scalar
            or concrete_type.is_enum
        )
        ret = concrete_type  # currently there is no need for a "proxied" type.

    elif obj_type := concrete_type.is_object_type:
        ret = _evaluate_object_type(
            type_info=type_info,
            concrete=obj_type,
            selection_set=selection_set_node,
            path=path,
        )
    elif lst := concrete_type.is_model:
        ret = _evaluate_list(
            type_info=type_info,
            concrete=lst,
            selection_set=selection_set_node,
            path=path,
        )
    elif interface := concrete_type.is_interface:
        ret = _evaluate_interface(
            type_info=type_info,
            concrete=interface,
            selection_set=selection_set_node,
            path=path,
        )
    elif is_union := concrete_type.is_union:
        ret = _evaluate_union(
            type_info=type_info,
            concrete=is_union,
            selection_set=selection_set_node,
            path=path,
        )
    if not ret:  # pragma: no cover
        raise NotImplementedError(f"type {concrete_type} not supported yet")

    if concrete_type.is_optional:
        return QtGqlOptional(of_type=ret)
    return ret


def _evaluate_field(
    type_info: OperationTypeInfo,
    concrete_field: QtGqlFieldDefinition,
    field_node: gql_lang.FieldNode,
    path: str,
    origin: QtGqlObjectType,
) -> QtGqlQueriedField:
    path += concrete_field.name
    return QtGqlQueriedField(
        type=_evaluate_selection_set_type(
            type_info,
            concrete_field.type,
            field_node.selection_set,
            path,
        ),
        concrete=concrete_field,
        variable_uses=_evaluate_variable_uses(type_info, concrete_field, field_node.arguments),
        origin=origin,
        type_info=type_info,
    )


def _evaluate_list(
    type_info: OperationTypeInfo,
    concrete: QtGqlList,
    selection_set: gql_lang.SelectionSetNode,
    path: str,
) -> QtGqlList:
    return QtGqlList(
        of_type=_evaluate_selection_set_type(
            type_info,
            concrete_type=concrete.of_type,
            selection_set_node=selection_set,
            path=path,
        ),
    )


def _evaluate_union(
    type_info: OperationTypeInfo,
    concrete: QtGqlUnion,
    selection_set: gql_lang.SelectionSetNode,
    path: str,
) -> QtGqlQueriedUnion:
    choices: dict[str, QtGqlQueriedObjectType] = {}
    if not has_typename_selection(selection_set):
        inject_typename_selection(selection_set)
    for selection in selection_set.selections:
        if is_field_node(selection):
            continue  # __typename selection
        fragment = is_inline_fragment(selection)
        assert fragment
        type_name = fragment.type_condition.name.value
        # unions support only object types http://spec.graphql.org/October2021/#sec-Unions
        resolved_type = require(concrete.get_by_name(type_name))
        choices[type_name] = _evaluate_object_type(
            type_info=type_info,
            concrete=resolved_type,
            selection_set=fragment.selection_set,
            path=path,
        )

    return QtGqlQueriedUnion(
        concrete=concrete,
        choices=tuple(choices.values()),
    )


def _unwrap_interface_fragments(
    type_info: SchemaTypeInfo,
    parent_concrete: QtGqlObjectType | QtGqlInterface,
    selection_set: gql_lang.SelectionSetNode,
    initial: dict[str, list[gql_lang.FieldNode]],
    id_was_selected: bool = False,
) -> dict[str, list[gql_lang.FieldNode]]:
    """Fragments can be nested, i.e node{ id.

        ...on SomeFrag{
            field1
            ...on Obj{
                name
            }
        }
    }
    :returns: all of the inline fragments as a flat list.
    """
    fields_for_concrete: list[gql_lang.FieldNode] = []
    if not has_id_selection(selection_set) and not id_was_selected:
        id_was_selected = True
        inject_id_selection(selection_set)
    for field_or_frag in selection_set.selections:
        if inline_frag := is_inline_fragment(field_or_frag):
            type_name = inline_frag.type_condition.name.value
            # no need to validate inner types are implementation, graphql-core does this.
            concrete_choice = type_info.get_object_type(
                type_name,
            ) or type_info.get_interface(type_name)
            assert concrete_choice
            _unwrap_interface_fragments(
                type_info,
                concrete_choice,
                inline_frag.selection_set,
                initial,
                id_was_selected,
            )
        else:
            field_node = is_field_node(field_or_frag)
            assert field_node
            if field_node.name.value != "__typename":
                fields_for_concrete.append(field_node)

    initial[parent_concrete.name] = fields_for_concrete
    return initial


def _evaluate_interface(
    type_info: OperationTypeInfo,
    concrete: QtGqlInterface,
    selection_set: gql_lang.SelectionSetNode,
    path: str,  # current path in the query tree.
) -> QtGqlQueriedInterface:
    choices: list[QtGqlQueriedObjectType] = []

    raw_fields_map = _unwrap_interface_fragments(
        type_info.schema_type_info,
        concrete,
        selection_set,
        {},
    )
    # dispatch fragmented fields where they are needed.
    for resolve_able in concrete.implementations.values():
        if concrete_choice := resolve_able.is_object_type:
            fields_for_obj = []
            # collect fields from bases.
            for base in concrete_choice.interfaces_raw:
                if fields_for_base := raw_fields_map.get(base.name, None):
                    fields_for_obj.extend(fields_for_base)

            # append fields of the choice itself
            if choice_fields := raw_fields_map.get(concrete_choice.name, None):
                fields_for_obj.extend(choice_fields)
            ss = gql_lang.SelectionSetNode(selections=tuple(fields_for_obj))
            choices.append(
                # This could probably be more optimized though, currently
                # this would suffice to reduce complexity.
                _evaluate_object_type(
                    type_info=type_info,
                    concrete=concrete_choice,
                    selection_set=ss,
                    path=path,
                ),
            )

    # inject __type_name selection, we'll use this to deserialize correctly.
    if not has_typename_selection(selection_set):
        inject_typename_selection(selection_set)

    name = f"{concrete.name}__{path}"
    ret = QtGqlQueriedInterface(
        name=name,
        concrete=concrete,
        choices=choices,
        fields_dict={
            field.name.value: _evaluate_field(
                type_info=type_info,
                concrete_field=concrete.fields_dict[field.name.value],
                path=path,
                field_node=field,
                origin=concrete,
            )
            for field in raw_fields_map[concrete.name]
        },
    )
    for choice in choices:
        choice.base_interface = ret
    type_info.narrowed_interfaces_map[name] = ret
    return ret


def _evaluate_object_type(
    type_info: OperationTypeInfo,
    concrete: QtGqlObjectType,
    selection_set: gql_lang.SelectionSetNode,
    path: str,  # current path in the query tree.
) -> QtGqlQueriedObjectType:
    # inject id selection for node implementors, it is required for caching purposes.
    if concrete.implements_node and not has_id_selection(selection_set):
        inject_id_selection(selection_set)

    fields: dict[str, QtGqlQueriedField] = {}
    for selection in selection_set.selections:
        if f_node := is_field_node(selection):
            if is_type_name_selection(f_node):
                continue  # __typename selection is handled with special care.
            concrete_field = concrete.fields_dict[f_node.name.value]
            fields[concrete_field.name] = _evaluate_field(
                type_info=type_info,
                concrete_field=concrete_field,
                field_node=f_node,
                path=path,
                origin=concrete,
            )

    name = f"{concrete.name}__{path}"
    if ret := type_info.narrowed_types_map.get(name, None):
        return ret

    ret = QtGqlQueriedObjectType(
        name=name,
        concrete=concrete,
        fields_dict=fields,
    )
    type_info.narrowed_types_map[name] = ret
    return ret


def _evaluate_operation(
    operation: OperationDefinitionNode,
    schema_type_info: SchemaTypeInfo,
) -> QtGqlOperationDefinition:
    """Each operation generates a whole new "proxy" schema. That schema will
    contain only the fields that are currently queried. The way we do that is
    creating a so-called "proxy objects".

    - Each proxy object mirrors a concrete object at schema level.
    - Each proxy object contains only the fields that was queried for this field in the tree.

    And because of that, one object type (at the concrete schema) might have many proxy objects.
    """
    type_info = OperationTypeInfo(schema_type_info)

    # input variables
    if variables_def := operation.variable_definitions:
        for var in variables_def:
            type_info.variables.append(_evaluate_variable(type_info.schema_type_info, var))

    selections = operation.selection_set
    root_type = type_info.schema_type_info.operation_types[operation.operation.value]
    assert root_type, f"Make sure you have {operation.operation.name} type defined in your schema"
    root_proxy_type = _evaluate_object_type(
        type_info=type_info,
        concrete=root_type,
        selection_set=selections,
        path="",
    )
    assert len(root_proxy_type.fields) == 1
    root_field = root_proxy_type.fields[0]
    return QtGqlOperationDefinition(
        root_field=root_field,
        root_type=root_proxy_type,
        operation_def=operation,
        variables=type_info.variables,
        narrowed_types=tuple(type_info.narrowed_types_map.values()),
        interfaces=tuple(type_info.narrowed_interfaces_map.values()),
    )


class _OperationsVisitor(visitor.Visitor):
    def __init__(self, type_info: SchemaTypeInfo):
        super().__init__()
        self.schema_type_info = type_info
        self.operations: dict[str, QtGqlOperationDefinition] = {}

    def enter_operation_definition(self, node, key, parent, path, ancestors) -> None:
        if operation := is_operation_def_node(node):
            if operation.operation in (
                OperationType.QUERY,
                OperationType.MUTATION,
                OperationType.SUBSCRIPTION,
            ):
                assert operation.name, "QtGql enforces operations to have names."
                self.operations[operation.name.value] = _evaluate_operation(
                    operation,
                    self.schema_type_info,
                )


def evaluate_operations(
    operations_document: graphql.DocumentNode,
    type_info: SchemaTypeInfo,
) -> dict[str, QtGqlOperationDefinition]:
    operation_visitor = _OperationsVisitor(type_info)
    graphql.visit(operations_document, operation_visitor)
    assert operation_visitor.operations
    return operation_visitor.operations
